from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from pyranges.core import names
from pyranges.core.example_data import example_data
from pyranges.core.pyranges_helpers import ensure_pyranges

if TYPE_CHECKING:
    import pyranges as pr


def random(
    n: int = 1000,
    length: int = 100,
    chromsizes: dict[str, int] | pd.DataFrame | None = None,
    seed: int | None = None,
    *,
    strand: bool = True,
) -> "pr.PyRanges":
    """Return PyRanges with random intervals.

    Parameters
    ----------
    n : int, default 1000
        Number of intervals.

    length : int, default 100
        Length of intervals.

    chromsizes : dict or DataFrame, default None, i.e. use "hg19"
        Draw intervals from within these bounds.

    strand : bool, default True
        Data should have strand.

    seed : int, default None
        Seed for random number generator.

    Examples
    --------
    >>> import pyranges as pr
    >>> pr.random(seed=12345)
    index    |    Chromosome    Start      End        Strand
    int64    |    object        int64      int64      object
    -------  ---  ------------  ---------  ---------  --------
    0        |    chr4          36129012   36129112   +
    1        |    chr5          177668498  177668598  -
    2        |    chr15         1902279    1902379    +
    3        |    chr11         23816613   23816713   +
    ...      |    ...           ...        ...        ...
    996      |    chr2          155410960  155411060  -
    997      |    chr6          80054552   80054652   +
    998      |    chrX          66474125   66474225   +
    999      |    chr7          69941721   69941821   -
    PyRanges with 1000 rows, 4 columns, and 1 index columns.
    Contains 24 chromosomes and 2 strands.

    """
    rng = np.random.default_rng(seed=seed)

    if chromsizes is None:
        df = example_data.chromsizes
    elif isinstance(chromsizes, dict):
        df = pd.DataFrame({names.CHROM_COL: list(chromsizes.keys()), names.END_COL: list(chromsizes.values())})
    else:
        df = chromsizes

    # Probability of picking each chromosome proportional to its size
    p = df.End / df.End.sum()

    # Pick chromosomes (by picking rows from df)
    chosen = rng.choice(df.index, size=n, p=p)

    # Extract the chosen chromosomes and their lengths
    chrom_chosen = df.loc[chosen, names.CHROM_COL].to_numpy()
    end_chosen = df.loc[chosen, names.END_COL].to_numpy() - length

    # Generate random starts, ensuring they don't exceed (End - length)
    random_starts = (rng.random(n) * end_chosen).astype(int)

    # Construct final DataFrame
    random_df = pd.DataFrame(
        {names.CHROM_COL: chrom_chosen, names.START_COL: random_starts, "End": random_starts + length}
    )

    if strand:
        s = rng.choice(["+", "-"], size=n)
        random_df["Strand"] = s

    return ensure_pyranges(random_df.reset_index(drop=True))



# below: Class to manage PyRanges options
# purposedly not using a docstring to avoid it being included in the docs


class PyRangesOptions:
    def __init__(self) -> None:
        self.options_in_use = {
            "max_rows_to_show": (8, "the max number of rows to show in PyRanges repr"),
            "max_column_names_to_show": (
                3,
                "how many columns listed in PyRanges repr when not all fit the screen width",
            ),
            "console_width": (None, "console width, affecting PyRanges representation (None for auto)"),
        }
        self.options_default = self.options_in_use.copy()

    def set_option(self, name: str, value: int) -> None:
        """Set an option to a new value.

        Set one or more options.

        Run pyranges.options.display_options() to see available options and their current values.

        Parameters
        ----------
        name : str
            The name of the option to set.

        value : int
            The value to set the option to.

        Examples
        --------
        >>> import pyranges as pr
        >>> pr.options.set_option('max_rows_to_show', 8)

        """
        if name in self.options_in_use:
            self.options_in_use[name] = (value, self.options_in_use[name][1])
        else:
            msg = f"Option {name} not recognized."
            raise ValueError(msg)

    def get_option(self, name: str) -> int:
        """Get the value of an option.

        Parameters
        ----------
        name : str
            The name of the option to get.

        Returns
        -------
        int
            The value of the option.

        Examples
        --------
        >>> import pyranges as pr
        >>> pr.options.get_option("max_rows_to_show")
        8

        """
        if name not in self.options_in_use:
            msg = f"Option {name} not recognized."
            raise ValueError(msg)
        return self.options_in_use[name][0]

    def reset_options(self) -> None:
        """Reset all options to their default values.

        Examples
        --------
        >>> import pyranges as pr
        >>> pr.options.get_option('max_rows_to_show')
        8

        >>> pr.options.set_option('max_rows_to_show', 10)
        >>> pr.options.get_option('max_rows_to_show')
        10

        >>> pr.options.set_option('console_width', 120)
        >>> pr.options.get_option('console_width')
        120

        >>> pr.options.reset_options()
        >>> pr.options.get_option('max_rows_to_show')
        8

        """
        self.options_in_use = self.options_default.copy()

    def display_options(self) -> str:
        """Return a representation of the current options and their values.

        Examples
        --------
        # In the below, the console width has been set to 120 so that the doctests will return the same result no matter
        # the console width.
        >>> import pyranges as pr
        >>> print(pr.options.display_options())
        max_rows_to_show         :   8 (the max number of rows to show in PyRanges repr)
        max_column_names_to_show :   3 (how many columns listed in PyRanges repr when not all fit the screen width)
        console_width            : 120 (console width, affecting PyRanges representation (None for auto))

        """
        max_len_k = max(len(k) for k in self.options_in_use)
        max_len_v = max(len(str(v[0])) for v in self.options_in_use.values())

        return "\n".join(f"{k:<{max_len_k}} : {v[0]!s:>{max_len_v}} ({v[1]})" for k, v in self.options_in_use.items())

    def __repr__(self) -> str:
        return self.display_options()


option_manager = PyRangesOptions()



import abc
from typing import TYPE_CHECKING

import numpy as np

from pyranges.core.names import END_COL, RANGE_COLS, START_COL

if TYPE_CHECKING:
    import pandas as pd

    from pyranges import RangeFrame


class InvalidRangesReason:
    """Describe why a range is invalid."""

    def __init__(self, invalid_part: "pd.DataFrame") -> None:
        self._invalid_part = invalid_part

    @property
    def invalid_part(self) -> "pd.DataFrame":
        """Return the invalid part of the range."""
        return self._invalid_part

    @property
    @abc.abstractmethod
    def reason(self) -> str:
        """Return the reason why the range is invalid."""

    def __str__(self) -> str:
        max_indices_to_show = 3
        indices = (
            self.invalid_part.index
            if len(self.invalid_part) < max_indices_to_show
            else [*self.invalid_part.index[:3], "..."]
        )
        return f"{self.reason}. See indexes: {', '.join([str(x) for x in indices])}"

    def __repr__(self) -> str:
        return str(self)

    @staticmethod
    @abc.abstractmethod
    def check_and_possibly_return_invalid_part(df: "pd.DataFrame") -> "InvalidRangesReason | None":
        """Check if a rangeframe is invalid and return the rows that are invalid."""

    def to_dict(self) -> dict[str, "str | pd.DataFrame"]:
        """Return a dict representation."""
        return {"reason": self.reason, "invalid_part": self.invalid_part}

    @staticmethod
    def formatted_reasons_list(ranges: "RangeFrame") -> str:
        """Return a formatted list of reasons why the range is invalid."""
        import textwrap

        if reasons := InvalidRangesReason.is_invalid_ranges_reasons(ranges):
            return textwrap.indent("\n".join([str(r) for r in reasons]), prefix="  * ")
        return ""

    @staticmethod
    def is_invalid_ranges_reasons(df: "pd.DataFrame") -> list["InvalidRangesReason"] | None:
        """Check if a rangeframe is invalid and return the reason(s) why.

        Returns
        -------
            list of InvalidRangesReason or None

        Examples
        --------
        >>> import pandas as pd
        >>> from pyranges import RangeFrame
        >>> valid = RangeFrame({"Start": [1, 5], "End": [10, 20]})
        >>> valid.reasons_why_frame_is_invalid() is None
        True
        >>> invalid = RangeFrame({"Start": [-5, 1, 100, np.nan], "End": [3, -2, -10, -40]})
        >>> invalid
          index  |        Start      End
          int64  |      float64    int64
        -------  ---  ---------  -------
              0  |           -5        3
              1  |            1       -2
              2  |          100      -10
              3  |          nan      -40
        RangeFrame with 4 rows, 2 columns, and 1 index columns.
        Invalid ranges:
          * 2 intervals are empty or negative length (end <= start). See indexes: 1, 2
          * 4 starts or ends are < 0. See indexes: 0, 1, 2, ...
          * 1 starts or ends are nan. See indexes: 3

        """
        invalid_ranges_reasons: list[InvalidRangesReason] = [
            invalid_ranges_reason
            for invalid_ranges_reason_class in sorted(InvalidRangesReason.__subclasses__(), key=lambda x: x.__name__)
            if (invalid_ranges_reason := invalid_ranges_reason_class.check_and_possibly_return_invalid_part(df))
        ]
        return invalid_ranges_reasons or None


class StartsOrEndsMissingValues(InvalidRangesReason):
    @property
    def reason(self) -> str:  # noqa: D102
        return f"{self.invalid_part.shape[0]} starts or ends are nan"

    @staticmethod
    def check_and_possibly_return_invalid_part(df: "pd.DataFrame") -> "InvalidRangesReason | None":  # noqa: D102
        any_missing_values = df[RANGE_COLS].isna()
        if any_missing_values.any().any():
            return StartsOrEndsMissingValues(df.loc[any_missing_values.any(axis=1)])
        return None


class EmptyOrNegativeIntervals(InvalidRangesReason):
    @property
    def reason(self) -> str:  # noqa: D102
        return f"{self.invalid_part.shape[0]} intervals are empty or negative length (end <= start)"

    @staticmethod
    def check_and_possibly_return_invalid_part(df: "pd.DataFrame") -> "InvalidRangesReason | None":  # noqa: D102
        if np.any(empty_or_negative_filter := (df[START_COL] >= df[END_COL])):
            return EmptyOrNegativeIntervals(df.loc[empty_or_negative_filter])
        return None


class StartOrEndsNegative(InvalidRangesReason):
    @property
    def reason(self) -> str:  # noqa: D102
        return f"{self.invalid_part.shape[0]} starts or ends are < 0"

    @staticmethod
    def check_and_possibly_return_invalid_part(df: "pd.DataFrame") -> "InvalidRangesReason | None":  # noqa: D102
        if (negative_filter := (df[START_COL] < 0) | (df[END_COL] < 0)).any():
            return StartOrEndsNegative(df.loc[negative_filter])
        return None



import glob

import pytest


def main():
    rst_files = ["docs/tutorial.rst", *glob.glob("docs/how_to*.rst")]
    # Build the command with all .rst files
    # Set the root directory explicitly
    args = ["--rootdir=.", *rst_files]
    # Call pytest with the list of .rst files
    return pytest.main(args)


if __name__ == "__main__":
    import sys

    sys.exit(main())



import io
import sys
from contextlib import redirect_stdout
from functools import wraps

import pandas as pd

import pyranges as pr
from pyranges.core.pyranges_helpers import ensure_pyranges

try:
    import fire  # type: ignore[reportMissingImports]
except ImportError:
    fire = None


@wraps(pd.read_csv)
def read_csv(path: str, **kwargs) -> pr.PyRanges:
    """Read a CSV with pandas, then convert the resulting DataFrame into a PyRanges."""
    df = pd.read_csv(path, **kwargs)
    return ensure_pyranges(pr.PyRanges(df))


# 1) Available readers (no from_string)
READERS = {
    "read_bed": pr.read_bed,
    "read_gtf": pr.read_gtf,
    "read_gff3": pr.read_gff3,
    "read_bam": pr.read_bam,
    "read_bigwig": pr.read_bigwig,
    "read_csv": read_csv,
}


def show_usage() -> None:
    """Display help/usage for the pyranges CLI."""
    prog = "pyranger"
    sys.stdout.write(
        f"""
pyranger: read sequence interval data into pyranges and apply a chain of methods

Usage:
  {prog} reader <args> , [var=reader <args>]… , method <args> , …

  • The command line defines a pipeline of actions separated by " , " and starting with readers
  • The first reader loads the main PyRanges object
  • Every reader beyond the first *must* be named (e.g. b=read_bed b.bed)
  • Methods are invoked on the main object; others can be provided as arguments, e.g. intersect_overlaps b
  • The result replaces the main object in the main

Available readers:
  {", ".join(READERS)}

PyRanges methods:
   see https://pyranges1.readthedocs.io/en/latest/pyranges_objects.html

Examples:
  1. Load only:
       {prog} read_bed sample1.bed

  2. Load + inspect first 5 lines:
       {prog} read_bed sample1.bed , head 5

  3. intersect_overlaps two files:
       {prog} read_bed a.bed , other=read_bed b.bed , intersect_overlaps other

  4. Chain involving three files:
       {prog} read_bed a.bed , b=read_bed b.bed , c=read_bed c.bed , join_overlaps b , intersect_overlaps c

Tip:
  Append `--help` immediately after any reader or method to see its documentation

""".strip()
        + "\n",
    )


def cast_literal(x: str) -> str | int | float | bool:
    """Try int, float, bool, else return the original string."""
    lx = x.lower()
    if lx in ("true", "false"):
        return lx == "true"
    try:
        return int(x)
    except ValueError:
        pass
    try:
        return float(x)
    except ValueError:
        pass
    return x


def main() -> None:  # noqa: C901,PLR0912,PLR0915
    """Entry point for the pyranges CLI."""
    if fire is None:
        sys.stderr.write(
            "Error: the CLI entry-point requires the Python package 'fire'.\n"
            "Please install it with:\n  pip install pyranges1[cli]\n"
        )
        sys.exit(1)

    args = sys.argv[1:]
    if not args or args[0] in ("-h", "--help"):
        show_usage()
        sys.exit(0)

    # 2) Split on literal commas
    segments: list[list[str]] = []
    buf: list[str] = []
    for tok in args:
        if tok == ",":
            if not buf:
                sys.exit("Error: empty segment before comma")
            segments.append(buf)
            buf = []
        else:
            buf.append(tok)
    if buf:
        segments.append(buf)

    # 3) Process the first reader: must be unnamed
    seg0 = segments[0]
    head0 = seg0[0]
    if "=" in head0 or head0 not in READERS:
        sys.exit(f"Error: the first segment must be an unnamed reader ({','.join(READERS)})")
    reader_fn = READERS[head0]
    reader_args = seg0[1:]
    # help for first reader?
    if any(a in ("-h", "--help") for a in reader_args):
        fire.Fire(reader_fn, name=head0, command=["--help"])
        sys.exit(0)
    # invoke reader (suppress Fire's print)
    with redirect_stdout(io.StringIO()):
        pr_obj = fire.Fire(reader_fn, command=reader_args)
    registry = {"pr": pr_obj}

    # 4) Process any additional *reader* segments (they must be named)
    n_readers = 1
    for seg in segments[1:]:
        head = seg[0]
        if "=" in head:
            var, cmd = head.split("=", 1)
            if cmd not in READERS:
                break  # not a reader segment
            reader_fn = READERS[cmd]
            reader_args = seg[1:]
            if any(a in ("-h", "--help") for a in reader_args):
                fire.Fire(reader_fn, name=cmd, command=["--help"])
                sys.exit(0)
            with redirect_stdout(io.StringIO()):
                obj = fire.Fire(reader_fn, command=reader_args)
            registry[var] = obj
            n_readers += 1
        else:
            break

    # 5) The rest are methods on registry['pr']
    primary = registry["pr"]
    for seg in segments[n_readers:]:
        head = seg[0]
        method_args = seg[1:]

        # method help?
        if any(a in ("-h", "--help") for a in method_args):
            fn = getattr(primary, head, None)
            if fn is None:
                sys.exit(f"Error: unknown method '{head}'")
            fire.Fire(fn, name=head, command=["--help"])
            sys.exit(0)

        fn = getattr(primary, head, None)
        if fn is None:
            sys.exit(f"Error: unknown method '{head}' on PyRanges")

        # parse flags & positionals, substituting any var names
        pos: list[object] = []
        flags: dict[str, object] = {}
        i = 0
        while i < len(method_args):
            tok = method_args[i]
            if tok.startswith("--"):
                if "=" in tok:
                    k, v = tok[2:].split("=", 1)
                    flags[k] = cast_literal(v)
                    i += 1
                else:
                    k = tok[2:]
                    if i + 1 < len(method_args) and not method_args[i + 1].startswith("--"):
                        flags[k] = cast_literal(method_args[i + 1])
                        i += 2
                    else:
                        flags[k] = True
                        i += 1
            else:
                if tok in registry:
                    pos.append(registry[tok])
                else:
                    pos.append(cast_literal(tok))
                i += 1

        # call the method
        primary = fn(*pos, **flags)
        registry["pr"] = primary  # update pr to the latest

    # 6) Print the final result
    if primary is not None:
        sys.stdout.write(str(primary) + "\n")


if __name__ == "__main__":
    main()



import importlib
import inspect
import pkgutil
from collections.abc import Iterable
from pathlib import Path


# --------------------------------------------------------------------------- #
# helpers                                                                     #
# --------------------------------------------------------------------------- #
def _iter_rst(project_root: Path) -> Iterable[tuple[str, str]]:
    """Yield every ``*.rst`` under docs/doc/tutorials (if present)."""
    for folder in ("docs", "doc", "tutorials"):
        base = project_root / folder
        if not base.exists():
            continue
        for path in base.rglob("*.rst"):
            yield str(path.relative_to(project_root)), path.read_text(encoding="utf-8")


def _public_from_module(module) -> Iterable[tuple[str, object]]:
    """Yield `(name, obj)` for symbols *defined in* ``module`` (no re-exports)."""
    for name, obj in vars(module).items():
        if name.startswith("_"):
            continue
        if getattr(obj, "__module__", None) == module.__name__:
            yield name, obj


def _iter_public_members(modname: str) -> Iterable[tuple[str, object]]:
    """Yield public `(name, obj)` pairs for *modname* and selected sub-packages."""
    try:
        mod = importlib.import_module(modname)
    except ModuleNotFoundError:
        return

    # 1 top-level names in the requested module
    yield from _public_from_module(mod)

    # 2  walk sub-packages, but only inside the public namespaces we care about
    if not hasattr(mod, "__path__"):
        return

    prefix = f"{modname}."
    allowed_prefixes = ("pyranges.seqs", "pyranges.orfs")

    for _, subname, _ in pkgutil.walk_packages(mod.__path__, prefix):
        if not subname.startswith(allowed_prefixes):
            continue
        try:
            subm = importlib.import_module(subname)
        except ImportError:
            continue
        yield from _public_from_module(subm)


def _format_doc(sig: str, obj) -> str:
    """Format *obj*'s docstring under a reST header."""
    doc = inspect.getdoc(obj) or ""
    return f"{'#' * 100}\n{sig}\n{'#' * len(sig)}\n{doc}\n"


def _iter_class_methods(
    cls,
    *,
    include_df: bool = False,
) -> Iterable[tuple[str, object]]:
    """Yield (name, obj) for methods defined **on cls itself**.

    If *include_df* is False we skip attributes not present in ``cls.__dict__``,
    i.e. those inherited unchanged from pandas DataFrame.
    """
    for n, o in inspect.getmembers(cls):
        if n.startswith("_"):
            continue
        if not callable(o):
            continue
        if include_df or n in cls.__dict__:  # defined or overridden here
            yield n, o


def _export_docs(
    to_file: str | Path | None = None,
    *,
    include_df: bool = False,
) -> str | None:
    pkg_root = Path(__file__).resolve().parents[1]  # …/pyranges
    pieces: list[str] = []

    # 1. RST sources
    for relpath, txt in _iter_rst(pkg_root.parent):
        pieces.append(f".. *** {relpath} ***\n\n{txt}")

    # 2. public API docstrings
    from pyranges import PyRanges, RangeFrame

    class_map = {"PyRanges": PyRanges, "RangeFrame": RangeFrame}
    public_modules = ["pyranges", "pyranges.seqs", "pyranges.orfs"]

    pieces.append("\n\n====================\nPublic API docstrings\n====================")

    for cls_name, cls in class_map.items():
        pieces.append(_format_doc(f"class {cls.__module__}.{cls_name}", cls))
        for meth_name, meth in _iter_class_methods(cls, include_df=include_df):
            pieces.append(_format_doc(f"{cls_name}.{meth_name}()", meth))

    for mod in public_modules:
        for name, obj in _iter_public_members(mod):
            if inspect.isfunction(obj):
                pieces.append(_format_doc(f"function {obj.__module__}.{name}()", obj))

    # 3. output
    full_text = "\n\n".join(pieces)

    if to_file is not None:
        path = Path(to_file)
        path.write_text(full_text, encoding="utf-8")
        return None

    return full_text



def pytest_configure(config) -> None:
    import pyranges as pr

    pr.options.set_option("max_rows_to_show", 8)
    pr.options.set_option("max_column_names_to_show", 3)
    pr.options.set_option("console_width", 120)
    # making sure reset_options would bring these value back:
    pr.options.options_default = pr.options.options_in_use.copy()



# import pyranges as pr
# import pandas as pd

# import pytest


# def test_new_position():

#     gr = pr.data.chipseq()
#     gr2 = pr.data.chipseq_background()

#     ju = gr.join(gr2, new_pos="union", suffixes="_A _B".split()).df

#     j = gr.join(gr2, suffix="_B")
#     ju2 = j.new_position("union", suffixes=(" ", "_B")).df
#     ju2.columns = ju.columns  # check_names did not work

#     pd.testing.assert_frame_equal(ju, ju2, check_names=False)



import subprocess  # nosec
import tempfile
from io import StringIO

import numpy as np
import pandas as pd
import pytest
from hypothesis import (
    HealthCheck,
    given,
    reproduce_failure,  # noqa: F401
    settings,
)
from natsort import natsorted  # type: ignore

import pyranges as pr
from tests.helpers import assert_df_equal
from tests.property_based.hypothesis_helper import (
    deadline,
    df_data,
    dfs_min,
    dfs_min_with_id,
    max_examples,
    selector,
)

# if environ.get("TRAVIS"):
#     max_examples = 100
#     deadline = None
# else:
#     max_examples = 1000
#     deadline = None

merge_command = "bedtools merge -o first,count -c 6,1 {} -i <(sort -k1,1 -k2,2n {})"


@pytest.mark.bedtools
@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all(),
)
@given(gr=dfs_min())  # pylint: disable=no-value-for-parameter
def test_merge(gr, strand) -> None:
    bedtools_strand = {True: "-s", False: ""}[strand]

    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = f"{temp_dir}/f1.bed"
        gr.df.to_csv(f1, sep="\t", header=False, index=False)

        cmd = merge_command.format(bedtools_strand, f1)

        # ignoring bandit security warning. All strings created by test suite
        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()  # nosec  # nosec

        if not strand:
            bedtools_df = pd.read_csv(
                StringIO(result),
                sep="\t",
                header=None,
                usecols=[0, 1, 2, 4],
                names=["Chromosome", "Start", "End", "Count"],
                dtype={"Chromosome": "category"},
            )
        else:
            bedtools_df = pd.read_csv(
                StringIO(result),
                sep="\t",
                header=None,
                names=["Chromosome", "Start", "End", "Strand", "Count"],
                dtype={"Chromosome": "category"},
            )

    result = gr.merge_overlaps(use_strand=strand)

    if not bedtools_df.empty:
        # need to sort because bedtools sometimes gives the result in non-natsorted chromosome order!
        if result.strand_valid:
            assert_df_equal(
                result.df.sort_values(["Chromosome", "Start", "Strand"]),
                bedtools_df.sort_values(["Chromosome", "Start", "Strand"]),
            )
        else:
            assert_df_equal(
                result.df.sort_values(["Chromosome", "Start"]),
                bedtools_df.sort_values(["Chromosome", "Start"]),
            )
    else:
        assert bedtools_df.empty == result.df.empty


cluster_command = "bedtools cluster {} -i <(sort -k1,1 -k2,2n {})"


@pytest.mark.bedtools
@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all(),
)
@given(gr=dfs_min())  # pylint: disable=no-value-for-parameter
def test_cluster(gr, strand) -> None:
    bedtools_strand = {True: "-s", False: ""}[strand]

    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = f"{temp_dir}/f1.bed"
        gr.df.to_csv(f1, sep="\t", header=False, index=False)

        cmd = cluster_command.format(bedtools_strand, f1)

        # ignoring bandit security warning. All strings created by test suite
        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()  # nosec  # nosec

        bedtools_df = pd.read_csv(
            StringIO(result),
            sep="\t",
            header=None,
            names=["Chromosome", "Start", "End", "Name", "Score", "Strand", "Cluster"],
            dtype={"Chromosome": "category"},
        )

    result = gr.cluster_overlaps(use_strand=strand)

    if not bedtools_df.empty:
        # need to sort because bedtools sometimes gives the result in non-natsorted chromosome order!
        sort_values = ["Chromosome", "Start", "Strand"] if result.strand_valid else ["Chromosome", "Start"]

        result_df = result.df.sort_values(sort_values)
        bedtools_df = bedtools_df.sort_values(sort_values)

        cluster_ids = dict(
            zip(
                result_df.Cluster.drop_duplicates(),
                bedtools_df.Cluster.drop_duplicates(),
                strict=False,
            ),
        )

        # bedtools gives different cluster ids than pyranges
        result_df.Cluster = result_df.Cluster.replace(cluster_ids)

        bedtools_df.Cluster = bedtools_df.Cluster.astype("int32")
        assert_df_equal(result_df.drop("Cluster", axis=1), bedtools_df.drop("Cluster", axis=1))
    else:
        assert bedtools_df.empty == result.df.empty


@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all(),
)
@given(gr=dfs_min_with_id())  # pylint: disable=no-value-for-parameter
def test_cluster_by(gr, strand) -> None:
    result = gr.cluster_overlaps(use_strand=strand, by="ID").df
    df = gr.df

    groupby = ["Chromosome", "Strand", "ID"] if strand else ["Chromosome", "ID"]

    grs = []

    for _, gdf in natsorted(df.groupby(groupby)):
        grs.append(pr.PyRanges(gdf))

    clusters = [gr.cluster_overlaps(use_strand=strand) for gr in grs]
    i = 1
    new_clusters = []
    for c in clusters:
        c.Cluster = i
        i += 1
        new_clusters.append(c)

    expected = pr.concat(new_clusters).df
    expected.loc[:, "Cluster"] = expected.Cluster.astype(np.int32)
    # expected = expected.drop_duplicates()

    assert_df_equal(result.drop("Cluster", axis=1), expected.drop("Cluster", axis=1))


@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all(),
)
@given(gr=dfs_min_with_id())  # pylint: disable=no-value-for-parameter
def test_merge_by(gr, strand) -> None:
    result = gr.merge_overlaps(by="ID").df.drop("ID", axis=1)

    df = gr.df

    grs = []
    for _, gdf in df.groupby("ID"):
        grs.append(pr.PyRanges(gdf))

    expected = pr.concat([gr.merge_overlaps() for gr in grs]).df

    assert_df_equal(result, expected)


makewindows_command = "bedtools makewindows -w 10 -b <(sort -k1,1 -k2,2n {})"


@pytest.mark.bedtools
@settings(
    max_examples=max_examples,
    print_blob=True,
    deadline=deadline,
    suppress_health_check=HealthCheck.all(),
)
@given(gr=dfs_min())  # pylint: disable=no-value-for-parameter
# @reproduce_failure('5.5.4', b'AXicY2RgYGAEIzgAsRkBAFsABg==')
def test_windows(gr) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = f"{temp_dir}/f1.bed"
        gr.df.to_csv(f1, sep="\t", header=False, index=False)

        cmd = makewindows_command.format(f1)

        # ignoring bandit security warning. All strings created by test suite
        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()  # nosec  # nosec

        bedtools_df = pd.read_csv(
            StringIO(result),
            sep="\t",
            header=None,
            names=["Chromosome", "Start", "End"],
            dtype={"Chromosome": "category"},
        )

    result = gr.window_ranges(10)[["Chromosome", "Start", "End"]].remove_strand()

    if not bedtools_df.empty:
        assert_df_equal(result.df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty


@pytest.mark.parametrize("strand", [True, False])
@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all(),
)
@given(gr=df_data())  # pylint: disable=no-value-for-parameter
def test_init(gr, strand) -> None:
    c, s, e, strands = gr

    if strand:
        pr.PyRanges(chromosomes=c, starts=s, ends=e, strands=strands)
    else:
        pr.PyRanges(chromosomes=c, starts=s, ends=e)


chipseq = pr.example_data.chipseq()


@settings(
    max_examples=max_examples,
    deadline=deadline,
    suppress_health_check=HealthCheck.all(),
)
@given(selector=selector())  # pylint: disable=no-value-for-parameter
def test_getitem(selector) -> None:
    if len(selector) == 3:
        a, b, c = selector
        chipseq[a, b, c]
    elif len(selector) == 2:
        a, b = selector
        chipseq[a, b]
    elif len(selector) == 1:
        a = selector[0]
        chipseq[a]
    elif len(selector) == 0:
        pass
    else:
        msg = "Should never happen"
        raise Exception(msg)


@pytest.mark.bedtools
@settings(
    max_examples=max_examples,
    deadline=deadline,
    print_blob=True,
    suppress_health_check=HealthCheck.all(),
)
@given(gr=dfs_min())  # pylint: disable=no-value-for-parameter
def test_summary(gr) -> None:
    # merely testing that it does not error
    # contents are just (pandas) dataframe.describe()
    gr.summary()



from numpy import nan

import pyranges as pr


def test_join_issue_4_right() -> None:
    import numpy as np

    chromsizes = pr.example_data.chromsizes
    query_regions = pr.tile_genome(chromsizes, int(1e6))
    signal_data = pr.example_data.chipseq
    signal_data["Score"] = np.random.randint(0, 100, len(signal_data))

    query_regions.join_overlaps(signal_data)


def test_join_issue_8() -> None:
    gd = {
        "Chromosome": ["chr1", "chr1", "chr1", "chr1"],
        "Start": [157, 584, 731, 821],
        "End": [257, 684, 831, 921],
        "Strand": ["-", "-", "-", "-"],
    }
    md = {
        "Chromosome": ["chr1", "chr1", "chr1", "chr1"],
        "Start": [316, 793, 889, 795],
        "End": [416, 893, 989, 895],
        "Strand": ["+", "+", "+", "-"],
    }

    g = pr.PyRanges(gd)
    m = pr.PyRanges(md)

    j = m.join_overlaps(g)
    expected_result = pr.PyRanges(
        {
            "Chromosome": ["chr1", "chr1"],
            "Start": [795, 795],
            "End": [895, 895],
            "Strand": ["-", "-"],
            "Start_b": [731, 821],
            "End_b": [831, 921],
        },
        index=[0, 1],
    )

    assert j.reset_index(drop=True).equals(expected_result)


def test_join_issue_8_right() -> None:
    gd = {
        "Chromosome": ["chr1", "chr1", "chr1", "chr1"],
        "Start": [157, 584, 731, 821],
        "End": [257, 684, 831, 921],
        "Strand": ["-", "-", "-", "-"],
    }
    md = {
        "Chromosome": ["chr1", "chr1", "chr1", "chr1"],
        "Start": [316, 793, 889, 795],
        "End": [416, 893, 989, 895],
        "Strand": ["+", "+", "+", "-"],
    }

    g = pr.PyRanges(gd)
    m = pr.PyRanges(md)

    j = m.join_overlaps(g, join_type="right")

    expected_result = pr.PyRanges(
        {
            "Chromosome": ["chr1", "chr1", nan, nan],
            "Start": [795.0, 795.0, nan, nan],
            "End": [895.0, 895.0, nan, nan],
            "Strand": ["-", "-", nan, nan],
            "Start_b": [731, 821, 157, 584],
            "End_b": [831, 921, 257, 684],
        },
        index=[0, 1, 2, 3],
    )
    assert j.reset_index(drop=True).equals(expected_result)



import pandas as pd

import pyranges as pr


def test_group_cumsum() -> None:
    """Test that group_cumsum works correctly with sorted/unsorted objects."""
    p = pr.PyRanges(
        {
            "Chromosome": [
                "chr2", "chr3", "chr2", "chr3", "chr3",
                "chr1", "chr2", "chr3", "chr3", "chr2",
                "chr2", "chr1", "chr1", "chr2", "chr3",
                "chr1", "chr2", "chr1", "chr3", "chr2"
            ],
            "Start": [
                785520, 89970, 626749, 168799, 372056,
                520458, 800809, 34720, 167104, 804868,
                59764, 972819, 235131, 4508, 866290,
                835796, 694284, 65536, 844945, 705072
            ],
            "End": [
                795577, 97475, 635187, 175999, 374410,
                521540, 807322, 41109, 167881, 805508,
                64323, 978796, 240207, 9362, 870401,
                845344, 701742, 73390, 851624, 712783
            ],
            "Strand": [
                "+", "-", "+", "+", "+",
                "+", "-", "+", "-", "-",
                "+", "+", "-", "+", "+",
                "-", "-", "-", "-", "+"
            ],
            "Label": [
                "grp0", "grp9", "grp2", "grp5", "grp5",
                "grp1", "grp7", "grp5", "grp9", "grp8",
                "grp2", "grp1", "grp3", "grp2", "grp5",
                "grp3", "grp7", "grp6", "grp9", "grp0"
            ]
        }    )

    def process_to_compare(df: pd.DataFrame) -> pd.DataFrame:
        return df[['Label', 'cs_end']].sort_values('cs_end')

    pcs=p.group_cumsum(group_by='Label', cumsum_end_column='cs_end', cumsum_start_column='cs_start')

    pcs_sorted = p.sort_ranges().group_cumsum(group_by='Label', cumsum_end_column='cs_end', cumsum_start_column='cs_start')

    pcs_sorted_lab = p.sort_ranges('Label').group_cumsum(group_by='Label', cumsum_end_column='cs_end',
                                              cumsum_start_column='cs_start')

    pd.testing.assert_frame_equal(process_to_compare(pcs), process_to_compare(pcs_sorted))
    pd.testing.assert_frame_equal(process_to_compare(pcs), process_to_compare(pcs_sorted_lab))






from typing import TYPE_CHECKING

from pyranges.core.pyranges_helpers import ensure_rangeframe, factorize_binary

if TYPE_CHECKING:
    from pyranges.range_frame.range_frame import RangeFrame


def _complement_overlaps(
    df: "RangeFrame",
    df2: "RangeFrame",
    by: list[str],
    slack: int | None = None,
) -> "RangeFrame":
    import ruranges

    if df.empty:
        return df

    factorized, factorized2 = factorize_binary(df, df2, by)

    indices = ruranges.complement_overlaps(
        groups=factorized,  # type: ignore[arg-type]
        starts=df.Start.to_numpy(),
        ends=df.End.to_numpy(),
        groups2=factorized2,  # type: ignore[arg-type]
        starts2=df2.Start.to_numpy(),
        ends2=df2.End.to_numpy(),
        slack=slack or 0,
    )

    return ensure_rangeframe(df.take(indices))  # type: ignore[arg-type]



import contextlib
import os

import pyranges as pr


def test_tostring() -> None:
    with contextlib.redirect_stdout(open(os.devnull, "w+")):
        pass


def test_tostring_index_and_column_share_name() -> None:
    gr = pr.example_data.f1
    gr = gr.set_index("Chromosome", drop=False)
    with contextlib.redirect_stdout(open(os.devnull, "w+")):
        pass



import pandas as pd
import pyranges as pr

def test_map_to_global() -> None:
    """Test that map_to_global works correctly with sorted/unsorted objects."""

    def process_to_compare(df: pd.DataFrame) -> pd.DataFrame:
        return df.sort_ranges()

    rh = pr.example_data.rfam_hits
    rh = rh[['target_name', 'seq_from', 'seq_to', 'strand', 'query_name', 'mdl_from', 'mdl_to']]
    rh = rh.rename(columns={'target_name': 'Chromosome', 'seq_from': 'Start', 'seq_to': 'End', 'strand': 'Strand'})
    rh = pr.PyRanges(rh)
    rh['Start'] -= 1  # convert to 0-based coordinates
    rh.loc[rh.Strand == '-', ['Start', 'End']] = rh.loc[rh.Strand == '-', ['End', 'Start']].values
    gr = pr.example_data.ncbi_gff
    gre = (gr[gr.Feature == 'exon']).get_with_loc_columns('Parent')
    rh = rh.remove_nonloc_columns()

    rhg_def = rh.map_to_global(gre, global_on='Parent', keep_id=True, keep_loc=True)
    rhg_sl1 = rh.sort_ranges().map_to_global(gre, global_on='Parent', keep_id=True, keep_loc=True)
    rhg_sg1 = rh.map_to_global(gre.sort_ranges(), global_on='Parent', keep_id=True, keep_loc=True)
    rhg_sg2 = rh.map_to_global(gre.sort_ranges('Parent'), global_on='Parent', keep_id=True, keep_loc=True)

    assert process_to_compare(rhg_def).equals(process_to_compare(rhg_sl1))
    assert process_to_compare(rhg_def).equals(process_to_compare(rhg_sg1))
    assert process_to_compare(rhg_def).equals(process_to_compare(rhg_sg2))

    # gre data is bad!


import pandas as pd

import pyranges as pr


def test_spliced_subsequence_use_strand_false() -> None:
    p = pr.PyRanges(
        {
            "Chromosome": [1, 1, 2, 2, 3],
            "Strand": ["+", "+", "-", "-", "+"],
            "Start": [1, 40, 10, 70, 140],
            "End": [11, 60, 25, 80, 152],
            "transcript_id": ["t1", "t1", "t2", "t2", "t3"],
        },
    )

    result = p.slice_ranges(0, 5, use_strand=False, group_by="transcript_id")

    _expected_result = {
        "Chromosome": {0: 1, 2: 2, 4: 3},
        "Strand": {0: "+", 2: "-", 4: "+"},
        "Start": {
            0: 1,
            2: 10,
            4: 140,
        },
        "End": {0: 6, 2: 15, 4: 145},
        "transcript_id": {0: "t1", 2: "t2", 4: "t3"},
    }
    expected_result = pd.DataFrame(_expected_result)
    pd.testing.assert_frame_equal(result, expected_result)


def test_spliced_subsequence_without_transcript_id() -> None:
    p = pr.PyRanges(
        {
            "Chromosome": [1, 1, 2, 2, 3],
            "Strand": ["+", "+", "-", "-", "+"],
            "Start": [1, 40, 10, 70, 140],
            "End": [11, 60, 25, 80, 152],
            "transcript_id": ["t1", "t1", "t2", "t2", "t3"],
        },
    )

    result = p.slice_ranges(0, 5, use_strand=False)

    _expected_result = {
        "Chromosome": {0: 1, 1: 1, 2: 2, 3: 2, 4: 3},
        "Strand": {0: "+", 1: "+", 2: "-", 3: "-", 4: "+"},
        "Start": {0: 1, 1: 40, 2: 10, 3: 70, 4: 140},
        "End": {0: 6, 1: 45, 2: 15, 3: 75, 4: 145},
        "transcript_id": {0: "t1", 1: "t1", 2: "t2", 3: "t2", 4: "t3"},
    }

    expected_result = pd.DataFrame(_expected_result)
    pd.testing.assert_frame_equal(result, expected_result)



from io import StringIO

import numpy as np
import pandas as pd
import pytest

import pyranges as pr


def string_to_df(c):
    return pd.read_csv(
        StringIO(c),
        sep=r"\s+",
        dtype={
            "Chromosome": "category",
            "Start": np.int32,
            "End": np.int32,
            "Start_b": np.int32,
            "End_b": np.int32,
        },
    )


@pytest.fixture
def gr():
    return pr.PyRanges(chromosomes="chr1", starts=[1, 15, 200], ends=[10, 20, 2000])


@pytest.fixture
def gr2():
    return pr.PyRanges(chromosomes="chr1", starts=[11, 11, 20, 20, 50], ends=[16, 20, 21, 22, 100])


@pytest.fixture
def result_k_nearest_different():
    c = """Chromosome  Start   End  Start_b  End_b  Distance
chr1      1    10       11     16         2
chr1      1    10       11     20         2
chr1      1    10       20     21        11
chr1      1    10       20     22        11
chr1      1    10       50    100        41
chr1     15    20       11     20         0
chr1     15    20       11     16         0
chr1     15    20       20     21         1
chr1     15    20       20     22         1
chr1    200  2000       50    100      -101"""

    return string_to_df(c)


@pytest.fixture
def result_k_nearest_different2():
    c = """Chromosome  Start  End  Start_b  End_b  Distance
0        chr1     11   16       15     20         0
1        chr1     11   20       15     20         0
2        chr1     11   20        1     10        -2
3        chr1     20   21       15     20         1
4        chr1     20   21        1     10       -11
5        chr1     20   21      200   2000       180
6        chr1     20   22       15     20         1
7        chr1     20   22        1     10       -11
8        chr1     20   22      200   2000       179
9        chr1     50  100       15     20       -31
10       chr1     50  100        1     10       -41"""

    return string_to_df(c)


def test_k_nearest_different(result_k_nearest_different, result_k_nearest_different2, gr, gr2) -> None:
    k = [3, 2, 1]
    r = gr.k_nearest(gr2, ties="different", k=k).df

    pd.testing.assert_frame_equal(r, result_k_nearest_different)

    ks = [1, 2, 3, 4, 2]
    r2 = gr2.k_nearest(gr, ties="different", k=ks).df

    # gr2.k = ks

    pd.testing.assert_frame_equal(r2, result_k_nearest_different2)


@pytest.fixture
def result_k_nearest_first():
    c = """Chromosome  Start   End  Start_b  End_b  Distance
0       chr1      1    10       11     16         2
1       chr1      1    10       20     21        11
2       chr1      1    10       50    100        41
3       chr1     15    20       11     20         0
4       chr1     15    20       20     21         1
5       chr1    200  2000       50    100      -101"""

    return string_to_df(c)


@pytest.fixture
def result_k_nearest_first2():
    c = """Chromosome  Start  End  Start_b  End_b  Distance
0        chr1     11   16       15     20         0
1        chr1     11   20       15     20         0
2        chr1     11   20        1     10        -2
3        chr1     20   21       15     20         1
4        chr1     20   21        1     10       -11
5        chr1     20   21      200   2000       180
6        chr1     20   22       15     20         1
7        chr1     20   22        1     10       -11
8        chr1     20   22      200   2000       179
9        chr1     50  100       15     20       -31
10       chr1     50  100        1     10       -41
11       chr1     50  100      200   2000       101"""

    return string_to_df(c)


def test_k_nearest_first(result_k_nearest_first, result_k_nearest_first2, gr, gr2) -> None:  # result_k_nearest_first,
    k = [3, 2, 1]
    r = gr.k_nearest(gr2, ties="first", k=k).df

    # print(r)
    # print(result_k_nearest_first)
    # assert 0

    pd.testing.assert_frame_equal(r, result_k_nearest_first)

    r2 = gr2.k_nearest(gr, ties="first", k=list(range(1, 5 + 1))).df

    # print(r2)
    # print(result_k_nearest_first2)
    # raise

    pd.testing.assert_frame_equal(r2, result_k_nearest_first2)


@pytest.fixture
def result_k_nearest_last():
    c = """Chromosome  Start   End  Start_b  End_b  Distance
0       chr1      1    10       11     20         2
1       chr1      1    10       20     22        11
2       chr1      1    10       50    100        41
3       chr1     15    20       11     20         0
4       chr1     15    20       20     22         1
5       chr1    200  2000       50    100      -101"""

    return string_to_df(c)


@pytest.fixture
def result_k_nearest_last2():
    c = """Chromosome  Start  End  Start_b  End_b  Distance
0        chr1     11   16       15     20         0
1        chr1     11   20       15     20         0
2        chr1     11   20        1     10        -2
3        chr1     20   21       15     20         1
4        chr1     20   21        1     10       -11
5        chr1     20   21      200   2000       180
6        chr1     20   22       15     20         1
7        chr1     20   22        1     10       -11
8        chr1     20   22      200   2000       179
9        chr1     50  100       15     20       -31
10       chr1     50  100        1     10       -41
11       chr1     50  100      200   2000       101"""

    return string_to_df(c)


def test_k_nearest_last(result_k_nearest_last, result_k_nearest_last2, gr, gr2) -> None:  # result_k_nearest_last,
    # print(gr)
    # print(gr2)
    k = [3, 2, 1]
    r = gr.k_nearest(gr2, ties="last", k=k).df

    pd.testing.assert_frame_equal(r, result_k_nearest_last)

    r2 = gr2.k_nearest(gr, ties="last", k=[1, 2, 3, 4, 5]).df

    # print("results")

    pd.testing.assert_frame_equal(r2, result_k_nearest_last2)
    # assert 0



import pandas.testing
import pytest

import pyranges as pr


def test_split_chunks() -> None:
    df = pr.example_data.ensembl_gtf.get_with_loc_columns(["gene_id"])
    frame1, frame2 = pr.parallelism.split_df_into_chunks_without_splitting_groups(df, by=["gene_id"], nb_splits=2)
    assert len({*frame1["gene_id"]}) == 1
    assert len({*frame2["gene_id"]}) == 1
    assert len(frame1) + len(frame2) == len(df)


@pytest.mark.slow
def test_split_chunks_many() -> None:
    gr = pr.PyRanges(
        {"Chromosome": ["chr1"] * 1000, "Start": range(1000), "End": range(1, 1001), "GeneId": [*range(100)] * 10}
    )
    res = pr.parallelism.split_df_into_chunks_without_splitting_groups(gr, by=["GeneId"], nb_splits=37)
    assert sum(len(r) for r in res) == len(gr)
    seen_ids = set()
    for i, r in enumerate(res):
        new_ids = {*r["GeneId"].drop_duplicates()}
        assert new_ids.isdisjoint(seen_ids), i
        seen_ids.update(new_ids)


@pytest.mark.slow
def test_run_in_parallel() -> None:
    df = pr.example_data.ensembl_gtf.get_with_loc_columns(["gene_id"])
    dfs = [gdf for _, gdf in df.groupby("gene_id")]
    res = pr.parallelism.run_in_parallel(
        function=lambda x: x.merge_overlaps(by="gene_id", count_col="Counts"),
        dfs=dfs,
        nb_cpu=2,
    ).reset_index(drop=True)
    expected_result = df.merge_overlaps(by="gene_id", count_col="Counts")
    pandas.testing.assert_frame_equal(res, expected_result)



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from pyranges.core.names import (
    CHROM_COL,
    END_COL,
    START_COL,
    STRAND_COL,
)
from pyranges.core.pyranges_helpers import arg_to_list

if TYPE_CHECKING:
    from pyranges import PyRanges

cumsum_start = "_refcumstart"
cumsum_end = "_refcumend"
suffix = "__ref"
start_b = START_COL + suffix
end_b = END_COL + suffix
strand_b = STRAND_COL + suffix

out_global_chromosome = "Chromosome_global"
out_global_start = "Start_global"
out_global_end = "End_global"
out_global_strand = "Strand_global"


def _map_to_local(
    gr,
    ref,
    ref_on,
    match_by,
    keep_chrom,
    keep_loc,
) -> "PyRanges":
    # record ordered columns of gr
    output_cols = gr.columns.tolist()
    gr_has_strand = gr.has_strand
    ref_has_strand = ref.has_strand
    match_by = arg_to_list(match_by)
    ref_cols_to_keep = [*match_by, ref_on]

    if ref_on in match_by:
        # dealing with case in which user used ref_on as match_by
        fixed_idcol = ref_on
        output_cols = [c for c in output_cols if c != ref_on]
        ref_cols_to_keep.pop(-1)  # remove idcol from match_by
    else:
        fixed_idcol = "__idcol"

    ref = (
        ref.get_with_loc_columns(ref_cols_to_keep)
        .group_cumsum(
            group_by=ref_on,
            use_strand="auto",
            cumsum_start_column=cumsum_start,
            cumsum_end_column=cumsum_end,
            keep_order=False,
        )
        .rename(columns={ref_on: fixed_idcol})
    )

    gr = gr.join_overlaps(ref, strand_behavior="ignore", match_by=match_by, suffix=suffix)

    # removing gr regions that do not overlap with ref
    gr = gr.combine_interval_columns("intersect", drop_old_columns=False, start2=start_b, end2=end_b)

    columns_to_drop = [cumsum_start, cumsum_end]
    if not keep_chrom:
        columns_to_drop = [*columns_to_drop, CHROM_COL]
    if not keep_loc:
        columns_to_drop = [*columns_to_drop, start_b, end_b]

    # dealing with - strand intervals on ref.
    if ref_has_strand:
        # depending on whether gr had strand, STRAND_COL + suffix may be missing in join_overlaps output
        strand_col_of_ref = strand_b if gr_has_strand else STRAND_COL

        ref_strand_is_neg = gr[strand_col_of_ref] == "-"

        # computing start and end for neg intervals at the same time, to allow replacement in one go
        gr.loc[ref_strand_is_neg, START_COL], gr.loc[ref_strand_is_neg, END_COL] = (
            (
                gr.loc[ref_strand_is_neg, end_b]
                - gr.loc[ref_strand_is_neg, END_COL]
                + gr.loc[ref_strand_is_neg, cumsum_start]
            ),
            (
                gr.loc[ref_strand_is_neg, end_b]
                - gr.loc[ref_strand_is_neg, START_COL]
                + gr.loc[ref_strand_is_neg, cumsum_start]
            ),
        )

        if gr_has_strand:  # and ref_has_strand: # implicit
            # transform Strand: for rows with Strand == Strand_b, it becomes '+', otherwise '-'
            gr[STRAND_COL] = np.where(gr[STRAND_COL] == gr[strand_b], "+", "-")
            if not keep_loc:
                columns_to_drop = [*columns_to_drop, strand_b]
    else:
        ref_strand_is_neg = pd.Series(data=False, index=gr.index, dtype=np.bool_)

    # dealing with + strand intervals in ref
    gr.loc[~ref_strand_is_neg, START_COL] = (
        gr.loc[~ref_strand_is_neg, START_COL]
        - gr.loc[~ref_strand_is_neg, start_b]
        + gr.loc[~ref_strand_is_neg, cumsum_start]
    )
    gr.loc[~ref_strand_is_neg, END_COL] = (
        gr.loc[~ref_strand_is_neg, END_COL]
        - gr.loc[~ref_strand_is_neg, start_b]
        + gr.loc[~ref_strand_is_neg, cumsum_start]
    )

    # drop useless columns

    if keep_chrom:
        gr = gr.rename(columns={CHROM_COL: out_global_chromosome})
        output_cols = [*output_cols, out_global_chromosome]

    gr = gr.drop(columns=columns_to_drop).rename(columns={fixed_idcol: CHROM_COL})

    if keep_loc:
        gr = gr.rename(columns={start_b: out_global_start, end_b: out_global_end})
        output_cols = [*output_cols, out_global_start, out_global_end]
        if gr_has_strand and ref_has_strand:
            gr = gr.rename(columns={strand_b: out_global_strand})
            output_cols = [*output_cols, out_global_strand]

    # reordering columns to match original gr
    return gr.reindex(columns=output_cols)



import inspect
import warnings
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from numpy import ndarray

from pyranges.core.names import (
    CHROM_AND_STRAND_COLS,
    CHROM_COL,
    FORWARD_STRAND,
    RANGE_COLS,
    REVERSE_STRAND,
    STRAND_BEHAVIOR_AUTO,
    STRAND_BEHAVIOR_IGNORE,
    STRAND_BEHAVIOR_OPPOSITE,
    STRAND_BEHAVIOR_SAME,
    STRAND_COL,
    STRICT_STRAND_BEHAVIOR_TYPE,
    USE_STRAND_AUTO,
    VALID_BY_OPTIONS,
    VALID_BY_TYPES,
    VALID_STRAND_BEHAVIOR_OPTIONS,
    VALID_STRAND_BEHAVIOR_TYPE,
    VALID_USE_STRAND_OPTIONS,
    VALID_USE_STRAND_TYPE,
)

if TYPE_CHECKING:
    from pyranges import PyRanges
    from pyranges.range_frame.range_frame import RangeFrame


def factorize(
    df: "pd.DataFrame",
    by: VALID_BY_TYPES,
) -> ndarray:
    """Factorize a DataFrame based on specified grouping columns.

    This function assigns a unique integer (of type uint32) to each group in the DataFrame,
    where groups are defined by the columns specified in ``by``. If ``by`` is empty or evaluates
    to False, an array of zeros is returned.

    Parameters
    ----------
    df : pd.DataFrame
        The input DataFrame containing the data to be factorized.
    by : VALID_BY_TYPES
        Column name(s) or criteria used to group the DataFrame. If empty, no grouping is performed.

    Returns
    -------
    ndarray
        A NumPy array of type uint32 containing group labels for each row in the DataFrame.

    """
    if not by:
        return np.zeros(len(df), dtype=np.uint8)
    _by = arg_to_list(by)
    return df.groupby(_by).ngroup().to_numpy().astype(np.uint32)


def factorize_binary(
    df: "pd.DataFrame",
    df2: "pd.DataFrame",
    by: VALID_BY_TYPES,
) -> tuple[ndarray, ndarray]:
    """Factorize two DataFrames simultaneously based on specified grouping columns.

    This function concatenates the specified columns from both DataFrames, assigns consistent group labels
    across the combined data, and then splits the labels back into two arrays corresponding to each DataFrame.
    If ``by`` is empty or evaluates to False, arrays of zeros are returned for both DataFrames.

    Parameters
    ----------
    df : pd.DataFrame
        The first DataFrame to be factorized.
    df2 : pd.DataFrame
        The second DataFrame to be factorized.
    by : VALID_BY_TYPES
        Column name(s) or criteria used to group both DataFrames.

    Returns
    -------
    tuple of ndarray
        A tuple containing two NumPy arrays of type uint32. The first array corresponds to group labels for
        ``df`` and the second array corresponds to group labels for ``df2``.

    """
    if not by:
        return np.zeros(len(df), dtype=np.uint8), np.zeros(len(df2), dtype=np.uint8)
    _by = arg_to_list(by)
    factorized = (
        pd.concat([df[_by], df2[_by]], ignore_index=True)
        .groupby(_by, observed=False)
        .ngroup()
        .astype(np.uint32)
        .to_numpy()
    )
    return factorized[: len(df)], factorized[len(df) :]


def split_on_strand(gr: "PyRanges") -> tuple["PyRanges", "PyRanges"]:
    """Split a PyRanges object into forward and reverse strand subsets.

    This function filters the input PyRanges object based on strand information, returning two PyRanges objects:
    one for the forward strand and another for the reverse strand.

    Parameters
    ----------
    gr : PyRanges
        A PyRanges object containing genomic ranges with strand information.

    Returns
    -------
    tuple of PyRanges
        A tuple containing two PyRanges objects: the first for ranges on the forward strand and the second for ranges
        on the reverse strand.

    """
    return (
        ensure_pyranges(gr.query(f"{STRAND_COL} == '{FORWARD_STRAND}'")),
        ensure_pyranges(gr.query(f"{STRAND_COL} == '{REVERSE_STRAND}'")),
    )


def prepare_by_single(
    self: "PyRanges",
    use_strand: VALID_USE_STRAND_TYPE,
    match_by: VALID_BY_OPTIONS,
) -> list[str]:
    """Prepare a list of column names for grouping or matching within a PyRanges object.

    This function determines which columns to use for operations by including the chromosome column by default
    and adding strand information if indicated by the ``use_strand`` parameter. Additional columns specified in
    ``match_by`` are appended to this list.

    Parameters
    ----------
    self : PyRanges
        The PyRanges object on which the operation is performed.
    use_strand : VALID_USE_STRAND_TYPE
        Indicator determining whether strand information should be included.
    match_by : VALID_BY_OPTIONS
        Additional column name(s) to use for grouping or matching.

    Returns
    -------
    list of str
        A list of column names used for grouping or matching.

    """
    use_strand = validate_and_convert_use_strand(self, use_strand)
    by = arg_to_list(match_by)
    return [*([CHROM_COL] if not use_strand else CHROM_AND_STRAND_COLS), *by]


def prepare_by_binary(
    self: "PyRanges",
    other: "PyRanges",
    strand_behavior: VALID_STRAND_BEHAVIOR_TYPE = "auto",
    match_by: VALID_BY_TYPES = None,
) -> tuple["PyRanges", list[str]]:
    """Prepare a secondary PyRanges object and determine the matching columns for binary operations.

    This function constructs a list of columns for matching (including chromosome and, optionally, strand columns)
    and may modify the secondary PyRanges object based on the specified strand behavior. For example, if the
    strand behavior is set to 'opposite', the strand information in the secondary object is inverted.

    Parameters
    ----------
    self : PyRanges
        The primary PyRanges object.
    other : PyRanges
        The secondary PyRanges object to be prepared.
    strand_behavior : VALID_STRAND_BEHAVIOR_TYPE, default "auto"
        Specifies how strand information should be handled. Options include 'auto', 'ignore', or 'opposite'.
    match_by : VALID_BY_TYPES, optional
        Additional column name(s) to include for matching.

    Returns
    -------
    tuple
        A tuple where the first element is the (possibly modified) secondary PyRanges object and the second element
        is a list of column names used for matching.

    """
    strand_behavior = validate_and_convert_strand_behavior(self, other, strand_behavior)
    default_cols = [CHROM_COL] if strand_behavior == STRAND_BEHAVIOR_IGNORE else CHROM_AND_STRAND_COLS
    by = [*default_cols, *arg_to_list(match_by)]

    if strand_behavior == STRAND_BEHAVIOR_OPPOSITE:
        _other = other.loc[:, [*RANGE_COLS, *by]].copy()
        _other.loc[:, STRAND_COL] = other[STRAND_COL].replace({"+": "-", "-": "+"})
    else:
        _other = other
    return ensure_pyranges(_other), by


def validate_and_convert_use_strand(self: "PyRanges", use_strand: VALID_USE_STRAND_TYPE) -> bool:
    """Validate and convert strand option.

    Used inside various methods accepting a PyRanges as input. It processes use_strand option provided by the user upon
    calling the function. This function validates the option and converts it to a boolean value.

    If use_strand is True, an error will be raised if the Strand column is missing or invalid.
    If use_strand is 'auto', it will be converted to True if the Strand column is valid, otherwise False.
    Note that if use_strand is 'auto' and the Strand column is invalid, a warning will be shown.

    Parameters
    ----------
    self : PyRanges
        PyRanges object to be validated.

    use_strand : {'auto', True, False}
        The use_strand option provided by the user.


    Returns
    -------
    bool
        The converted boolean value of use_strand.

    """
    if use_strand not in VALID_USE_STRAND_OPTIONS:
        msg = f"Invalid use_strand option: {use_strand}"
        raise ValueError(msg)
    if use_strand is True and not self.has_strand:
        msg = "Cannot have use_strand=True when the Strand column is missing."
        raise ValueError(msg)
    if use_strand == USE_STRAND_AUTO:
        use_strand = self.strand_valid

        # If Strand column is invalid, set use_strand to False
        # but if there are intervals on the negative strand, warn user
        if not use_strand and contains_negative_strand(self):
            fn_name = stack[2].function if len(stack := inspect.stack()) > 2 else "unknown_function"  # noqa: PLR2004
            fn_name = stack[1].function if fn_name == "<module>" else fn_name
            msg = f"{fn_name}: '{USE_STRAND_AUTO}' use_strand treated as False due to invalid Strand values. Please use use_strand=False"
            warnings.warn(msg, stacklevel=4)
    return use_strand


def validate_and_convert_strand_behavior(
    self: "PyRanges",
    other: "PyRanges",
    strand_behavior: VALID_STRAND_BEHAVIOR_TYPE,
) -> STRICT_STRAND_BEHAVIOR_TYPE:
    """Validate and convert strand behavior.

    Used inside various methods accepting two PyRanges as input. It processes strand_behavior option provided
    by the user upon calling the function. This function validates the option and converts it to a valid unambiguous
    value.
    """
    if strand_behavior not in VALID_STRAND_BEHAVIOR_OPTIONS:
        msg = f"{VALID_STRAND_BEHAVIOR_OPTIONS} are the only valid values for strand_behavior. Was: {strand_behavior}"
        raise ValueError(msg)

    if not (self.strand_valid and other.strand_valid):
        if strand_behavior == STRAND_BEHAVIOR_AUTO:
            # If any Strand column is invalid, set strand_behavior to STRAND_BEHAVIOR_IGNORE
            # but warn user if both objects are Stranded and there are intervals on the negative strand
            strand_behavior = STRAND_BEHAVIOR_IGNORE
            if (
                self.has_strand
                and other.has_strand
                and (contains_negative_strand(self) or contains_negative_strand(other))
            ):
                fn_name = stack[2].function if len(stack := inspect.stack()) > 2 else "unknown_function"  # noqa: PLR2004
                fn_name = stack[1].function if fn_name == "<module>" else fn_name
                msg = (
                    f"{fn_name}: '{STRAND_BEHAVIOR_AUTO}' strand_behavior treated as {STRAND_BEHAVIOR_IGNORE} due to invalid Strand values. "
                    f"Please use strand_behavior={STRAND_BEHAVIOR_IGNORE}"
                )
                warnings.warn(msg, stacklevel=4)

        elif strand_behavior in (STRAND_BEHAVIOR_SAME, STRAND_BEHAVIOR_OPPOSITE):
            msg = f"Can only do {strand_behavior} strand operations when both PyRanges contain valid strand info."
            raise ValueError(msg)
    elif strand_behavior == STRAND_BEHAVIOR_AUTO:
        strand_behavior = STRAND_BEHAVIOR_SAME

    return strand_behavior


def group_keys_from_validated_strand_behavior(
    strand_behavior: VALID_STRAND_BEHAVIOR_TYPE,
    by: VALID_BY_OPTIONS = None,
) -> list[str]:
    """Return group keys based on strand behavior.

    If strand_behavior is 'same', return [CHROM_COL, STRAND_COL].
    If strand_behavior is 'opposite', return [CHROM_COL, STRAND_COL].
    If strand_behavior is 'ignore', return [CHROM_COL].

    In each case, if a by argument is provided, it is appended/extended to the list of group keys.

    """
    if strand_behavior == STRAND_BEHAVIOR_AUTO:
        msg = "this function must be called with a validated strand_behavior"
        raise ValueError(msg)

    strand_cols = [STRAND_COL] if strand_behavior in [STRAND_BEHAVIOR_SAME, STRAND_BEHAVIOR_OPPOSITE] else []

    return [CHROM_COL, *strand_cols, *arg_to_list(by)]


def strand_behavior_from_validated_use_strand(
    df: "PyRanges",
    use_strand: VALID_USE_STRAND_TYPE,
) -> VALID_STRAND_BEHAVIOR_TYPE:
    """Return strand_behavior based on use_strand bool.

    If strand is True, returns 'same' strand behavior, otherwise 'ignore' strand behavior.

    This function must be used with use_strand validated and converted to a boolean value.
    """
    return STRAND_BEHAVIOR_SAME if validate_and_convert_use_strand(df, use_strand) else STRAND_BEHAVIOR_IGNORE


def use_strand_from_validated_strand_behavior(
    self: "PyRanges",
    other: "PyRanges",
    strand_behavior: VALID_STRAND_BEHAVIOR_TYPE,
) -> bool:
    """Return use_strand based on strand_behavior.

    If strand_behavior is 'ignore', returns False, otherwise True if both PyRanges contain valid strand info.

    This function must be used with strand_behavior validated and converted to either 'same', 'opposite' or 'ignore'.
    """
    strand_behavior = validate_and_convert_strand_behavior(self, other, strand_behavior)
    return strand_behavior != STRAND_BEHAVIOR_IGNORE


def ensure_pyranges(df: "pd.DataFrame") -> "PyRanges":
    """Ensure df is a PyRanges.

    Helps pyright.
    """
    from pyranges import PyRanges

    if not isinstance(ret := PyRanges(df, copy=False), PyRanges):
        msg = "Not a PyRanges"
        raise TypeError(msg)
    return ret


def ensure_rangeframe(df: "pd.DataFrame") -> "RangeFrame":
    """Ensure df is a rangeframe.

    Helps pyright.
    """
    from pyranges.range_frame.range_frame import RangeFrame

    if not isinstance(ret := RangeFrame(df, copy=False), RangeFrame):
        msg = "Not a RangeFrame"
        raise TypeError(msg)
    return ret


def arg_to_list(by: VALID_BY_OPTIONS) -> list[str]:
    """Convert a by or column_key argument to a list.

    If by is a string, it will be converted to a list.
    If by is None, an empty list will be returned.
    If by is already a list, it will be returned as is.
    """
    return [by] if isinstance(by, str) else ([*by] if by is not None else [])


def contains_negative_strand(self: "PyRanges") -> bool:
    """Return True if negative strand is present."""
    return self.has_strand and REVERSE_STRAND in self[STRAND_COL].to_numpy()



import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pandas as pd
from natsort import natsorted  # type: ignore[import]

from pyranges.core.pyranges_helpers import ensure_pyranges

if TYPE_CHECKING:
    from collections.abc import Mapping

    from pyranges.core.pyranges_main import PyRanges

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def from_string(s: str) -> "PyRanges":
    """Create a PyRanges from multiline string.

    Parameters
    ----------
    s : str
        String with data.

    Examples
    --------
    >>> import pyranges as pr
    >>> s = '''Chromosome      Start        End Strand
    ... chr1  246719402  246719502      +
    ... chr5   15400908   15401008      +
    ... chr9   68366534   68366634      +
    ... chr14   79220091   79220191      +
    ... chr14  103456471  103456571      -'''

    >>> pr.from_string(s)
      index  |    Chromosome        Start        End  Strand
      int64  |    object            int64      int64  object
    -------  ---  ------------  ---------  ---------  --------
          0  |    chr1          246719402  246719502  +
          1  |    chr5           15400908   15401008  +
          2  |    chr9           68366534   68366634  +
          3  |    chr14          79220091   79220191  +
          4  |    chr14         103456471  103456571  -
    PyRanges with 5 rows, 4 columns, and 1 index columns.
    Contains 4 chromosomes and 2 strands.

    """
    from io import StringIO

    df = pd.read_csv(StringIO(s), sep=r"\s+", index_col=None)

    return ensure_pyranges(df)


def read_bed(f: Path, /, nrows: int | None = None) -> "PyRanges":
    """Return bed file as PyRanges.

    This is a reader for files that follow the bed format. They can have from
    3-12 columns which will be named like so:

    Chromosome Start End Name Score Strand ThickStart ThickEnd ItemRGB
    BlockCount BlockSizes BlockStarts

    Parameters
    ----------
    f : str
        Path to bed file

    nrows : Optional int, default None
        Number of rows to return.

    Notes
    -----
    If you just want to create a PyRanges from a tab-delimited bed-like file,
    use `pr.PyRanges(pandas.read_table(f))` instead.

    Returns
    -------
    PyRanges


    Examples
    --------
    >>> import pyranges as pr
    >>> path = pr.example_data.files["aorta.bed"]
    >>> pr.read_bed(path, nrows=5)
      index  |    Chromosome      Start      End  Name        Score  Strand
      int64  |    category        int64    int64  object      int64  category
    -------  ---  ------------  -------  -------  --------  -------  ----------
          0  |    chr1             9916    10115  H3K27me3        5  -
          1  |    chr1             9939    10138  H3K27me3        7  +
          2  |    chr1             9951    10150  H3K27me3        8  -
          3  |    chr1             9953    10152  H3K27me3        5  +
          4  |    chr1             9978    10177  H3K27me3        7  -
    PyRanges with 5 rows, 6 columns, and 1 index columns.
    Contains 1 chromosomes and 2 strands.

    """
    columns = [
        "Chromosome",
        "Start",
        "End",
        "Name",
        "Score",
        "Strand",
        "ThickStart",
        "ThickEnd",
        "ItemRGB",
        "BlockCount",
        "BlockSizes",
        "BlockStarts",
    ]
    path = Path(f)
    if path.name.endswith(".gz"):
        import gzip

        first_start = gzip.open(path).readline().decode().split()[1]  # noqa: SIM115
    else:
        first_start = path.open().readline().split()[1]

    header = None

    try:
        int(first_start)
    except ValueError:
        header = 0

    ncols = pd.read_table(path, nrows=2).shape[1]

    df = pd.read_csv(
        path,
        dtype={"Chromosome": "category", "Strand": "category"},
        nrows=nrows,
        header=header,
        names=columns[:ncols] if header != 0 else None,
        sep="\t",
    )

    df.columns = pd.Index(columns[: df.shape[1]])

    return ensure_pyranges(df)


def read_bam(
    f: str | Path,
    /,
    mapq: int = 0,
    required_flag: int = 0,
    filter_flag: int = 1540,
    *,
    sparse: bool = True,
) -> "PyRanges":
    """Return bam file as PyRanges.

    Parameters
    ----------
    f : str
        Path to bam file

    sparse : bool, default True
        Whether to return only the columns Chromosome, Start, End, Strand, Flag.
        Set to False to return also columns
        QueryStart, QueryEnd, QuerySequence, Name, Cigar, Quality (more time consuming).

    mapq : int, default 0
        Minimum mapping quality score.

    required_flag : int, default 0
        Flags which must be present for the interval to be read.

    filter_flag : int, default 1540
        Ignore reads with these flags. Default 1540, which means that either
        the read is unmapped, the read failed vendor or platfrom quality
        checks, or the read is a PCR or optical duplicate.

    Returns
    -------
    PyRanges


    Notes
    -----
    This functionality requires the library `bamread`. It can be installed with
    `pip install bamread` or `conda install -c bioconda bamread`.

    Examples
    --------
    >>> import pyranges as pr
    >>> path = pr.example_data.files["smaller.bam"]
    >>> pr.read_bam(path)
    index    |    Chromosome    Start     End       Strand      Flag
    int64    |    category      int64     int64     category    uint16
    -------  ---  ------------  --------  --------  ----------  --------
    0        |    chr1          887771    887796    -           16
    1        |    chr1          994660    994685    -           16
    2        |    chr1          1041102   1041127   +           0
    3        |    chr1          1770383   1770408   -           16
    ...      |    ...           ...       ...       ...         ...
    96       |    chr1          18800901  18800926  +           0
    97       |    chr1          18800901  18800926  +           0
    98       |    chr1          18855123  18855148  -           16
    99       |    chr1          19373470  19373495  +           0
    PyRanges with 100 rows, 5 columns, and 1 index columns.
    Contains 1 chromosomes and 2 strands.

    """
    path = Path(f)
    try:
        import bamread  # type: ignore[import]
    except ImportError:
        LOGGER.exception(
            "bamread must be installed to read bam. Use `conda install -c bioconda bamread` or `pip install bamread` to install it.",
        )
        sys.exit(1)

    if bamread.__version__ in {
        "0.0.1",
        "0.0.2",
        "0.0.3",
        "0.0.4",
        "0.0.5",
        "0.0.6",
        "0.0.7",
        "0.0.8",
        "0.0.9",
    }:
        LOGGER.exception(
            "bamread not recent enough. Must be 0.0.10 or higher. Use `conda install -c bioconda 'bamread>=0.0.10'` or `pip install bamread>=0.0.10` to install it.",
        )
        sys.exit(1)

    if sparse:
        return ensure_pyranges(bamread.read_bam(path, mapq, required_flag, filter_flag))
    df = bamread.read_bam_full(path, mapq, required_flag, filter_flag)
    return ensure_pyranges(df)


def _fetch_gene_transcript_exon_id(attribute: pd.Series, annotation: str | None = None) -> pd.DataFrame:
    no_quotes = attribute.str.replace('"', "").str.replace("'", "")

    df = no_quotes.str.extract(
        "gene_id.?(.+?);(?:.*group_by.?(.+?);)?(?:.*exon_number.?(.+?);)?(?:.*exon_id.?(.+?);)?",
        expand=True,
    )  # .iloc[:, [1, 2, 3]]

    df.columns = pd.Index(["gene_id", "group_by", "exon_number", "exon_id"])

    if annotation == "ensembl":
        newdfs = []
        for c in ["gene_id", "group_by", "exon_id"]:
            r = df[c].astype(str).str.extract(r"(\d+)").astype(float)
            newdfs.append(r)

        newdf = pd.concat(newdfs, axis=1)
        newdf.insert(2, "exon_number", df["exon_number"])
        df = newdf

    return df


def find_first_data_line_index(file_path: Path) -> int:
    """Find the first line that is not a comment."""
    first_non_comment = 0
    try:
        import gzip

        with gzip.open(file_path) as zh:
            for i, zl in enumerate(zh):
                if zl.decode()[0] != "#":
                    first_non_comment = i
                    break
    except (OSError, TypeError):  # not a gzipped file, or StringIO
        fh = file_path.open()
        for i, line in enumerate(fh):
            if line[0] != "#":
                first_non_comment = i
                break
        fh.close()

    return first_non_comment


def read_gtf(
    f: str | Path,
    /,
    nrows: bool | None = None,
    *,
    full: bool = True,
    duplicate_attr: bool = False,
    ignore_bad: bool = False,
) -> "PyRanges":
    r"""Read files in the Gene Transfer Format.

    Parameters
    ----------
    f : str
        Path to GTF file.

    full : bool, default True
        Whether to read and interpret the annotation column.

    nrows : int, default None
        Number of rows to read. Default None, i.e. all.

    duplicate_attr : bool, default False
        Whether to handle (potential) duplicate attributes or just keep last one.

    ignore_bad : bool, default False
        Whether to ignore bad lines or raise an error.


    Returns
    -------
    PyRanges

    Note
    ----
    The GTF format encodes both Start and End as 1-based included.
    PyRanges encodes intervals as 0-based, Start included and End excluded.

    See Also
    --------
    pyranges.read_gff3 : read files in the General Feature Format

    Examples
    --------
    >>> import pyranges as pr
    >>> from tempfile import NamedTemporaryFile
    >>> contents = ['#!genome-build GRCh38.p10']
    >>> contents.append('1\thavana\tgene\t11869\t14409\t.\t+\t.\tgene_id "ENSG00000223972"; gene_version "5"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene";')
    >>> contents.append('1\thavana\ttranscript\t11869\t14409\t.\t+\t.\tgene_id "ENSG00000223972"; gene_version "5"; group_by "ENST00000456328"; transcript_version "2"; gene_name "DDX11L1"; gene_source "havana"; gene_biotype "transcribed_unprocessed_pseudogene"; transcript_name "DDX11L1-202"; transcript_source "havana"; transcript_biotype "processed_transcript"; tag "basic"; transcript_support_level "1";')
    >>> f = NamedTemporaryFile("w")
    >>> _bytes_written = f.write("\n".join(contents))
    >>> f.flush()
    >>> pr.read_gtf(f.name)
      index  |      Chromosome  Source    Feature       Start      End  Score     Strand      Frame     gene_id          ...
      int64  |        category  object    category      int64    int64  object    category    object    object           ...
    -------  ---  ------------  --------  ----------  -------  -------  --------  ----------  --------  ---------------  -----
          0  |               1  havana    gene          11868    14409  .         +           .         ENSG00000223972  ...
          1  |               1  havana    transcript    11868    14409  .         +           .         ENSG00000223972  ...
    PyRanges with 2 rows, 20 columns, and 1 index columns. (11 columns not shown: "gene_version", "gene_name", "gene_source", ...).
    Contains 1 chromosomes and 1 strands.

    """
    path = Path(f)
    _skiprows = find_first_data_line_index(path)

    if full:
        gr = read_gtf_full(
            path,
            nrows=nrows,
            skiprows=_skiprows,
            duplicate_attr=duplicate_attr,
            ignore_bad=ignore_bad,
        )
    else:
        gr = read_gtf_restricted(path, _skiprows, nrows=None)

    return gr


def read_gtf_full(
    f: str | Path,
    /,
    nrows: int | None = None,
    skiprows: int = 0,
    chunksize: int = int(1e5),  # for unit-testing purposes
    *,
    duplicate_attr: bool = False,
    ignore_bad: bool = False,
) -> "PyRanges":
    """Read files in the Gene Transfer Format into a PyRanges, including the annotation column.

    Parameters
    ----------
    f : str
        Path to GTF file.

    nrows : int, default None
        Number of rows to read. Default None, i.e. all.

    skiprows : int, default 0
        Number of rows to skip. Default 0.

    chunksize : int, default 100000
        Number of rows to read at a time. Default 100000.

    duplicate_attr : bool, default False
        Whether to handle (potential) duplicate attributes or just keep last one.

    ignore_bad : bool, default False
        Whether to ignore bad lines or raise an error.

    Returns
    -------
    PyRanges

    """
    dtypes: Mapping = {"Chromosome": "category", "Feature": "category", "Strand": "category"}

    names = ["Chromosome", "Source", "Feature", "Start", "End", "Score", "Strand", "Frame", "Attribute"]
    path = Path(f)

    df_iter = pd.read_csv(
        path,
        sep="\t",
        header=None,
        names=names,
        dtype=dtypes,
        chunksize=chunksize,
        skiprows=skiprows,
        nrows=nrows,
    )

    _to_rows = to_rows_keep_duplicates if duplicate_attr else to_rows

    dfs = []
    for df in df_iter:
        extra = _to_rows(df.Attribute.astype(str), ignore_bad=ignore_bad)
        _df = df.drop("Attribute", axis=1)
        extra = extra.set_index(_df.index)
        ndf = pd.concat([_df, extra], axis=1, sort=False)
        dfs.append(ndf)

    df = pd.concat(dfs, sort=False)
    df.loc[:, "Start"] = df.Start - 1

    return ensure_pyranges(df)


def parse_kv_fields(line: str) -> list[list[str]]:
    """Parse GTF attribute column."""
    return [kv.replace('""', '"NA"').replace('"', "").split(None, 1) for kv in line.rstrip("; ").split("; ")]


def to_rows(anno: pd.Series, *, ignore_bad: bool = False) -> pd.DataFrame:
    """Parse GTF attribute column into a dataframe of attribute columns."""
    entry = ""
    try:
        row = anno.head(1)
        for entry in row:
            str(entry).replace('"', "").replace(";", "").split()
    except AttributeError:
        msg = f"Invalid attribute string: {entry}. If the file is in GFF3 format, use pr.read_gff3 instead."
        raise AttributeError(msg) from AttributeError

    rowdicts = []
    line = ""
    try:
        for line in anno:
            rowdicts.append(dict(parse_kv_fields(line)))
    except ValueError:
        if not ignore_bad:
            LOGGER.exception(
                "The following line is not parseable as gtf:\n%s\n\nTo ignore bad lines use ignore_bad=True.",
                line,
            )
            raise

    return pd.DataFrame.from_records(rowdicts)


def to_rows_keep_duplicates(anno: pd.Series, *, ignore_bad: bool = False) -> pd.DataFrame:
    """If an entry is found multiple times in the attribute string, keep all of them.

    Examples
    --------
    >>> anno = pd.Series(["gene DDX11L1; gene sonic; unique hi"])
    >>> result = to_rows_keep_duplicates(anno)
    >>> result.to_dict(orient="records")
    [{'gene': 'DDX11L1,sonic', 'unique': 'hi'}]

    """
    rowdicts = []
    line = ""
    try:
        for line in anno:
            rowdict = {}

            # rstrip: allows for GFF not having a last ";", or having final spaces
            for k, v in tuple(parse_kv_fields(line)):
                if k not in rowdict:
                    rowdict[k] = [v]
                else:
                    rowdict[k].append(v)

            rowdicts.append({k: ",".join(v) if isinstance(v, list) else v for k, v in rowdict.items()})
    except ValueError:
        if not ignore_bad:
            LOGGER.exception(
                "The following line is not parseable as gtf:\n\n%s\n\nTo ignore bad lines use ignore_bad=True.",
                line,
            )

    return pd.DataFrame.from_records(rowdicts)


def read_gtf_restricted(f: str | Path, skiprows: int | None, nrows: int | None = None) -> "PyRanges":
    """Read certain columns from GTF file.

    Seqname - name of the chromosome or scaffold; chromosome names can be given with or without the 'chr' prefix. Important note: the seqname must be one used within Ensembl, i.e. a standard chromosome name or an Ensembl identifier such as a scaffold ID, without any additional content such as species or assembly. See the example GFF output below.
    # source - name of the program that generated this feature, or the data source (database or project name)
    feature - feature type name, e.g. Gene, Variation, Similarity
    start - Start position of the feature, with sequence numbering starting at 1.
    end - End position of the feature, with sequence numbering starting at 1.
    score - A floating point value.
    strand - defined as + (forward) or - (reverse).
    # frame - One of '0', '1' or '2'. '0' indicates that the first base of the feature is the first base of a codon, '1' that the second base is the first base of a codon, and so on..
    attribute - A semicolon-separated list of tag-value pairs, providing additional information about each feature.

    """
    dtypes: Mapping = {"Chromosome": "category", "Feature": "category", "Strand": "category"}
    path = Path(f)

    df_iter = pd.read_csv(
        path,
        sep="\t",
        comment="#",
        usecols=[0, 2, 3, 4, 5, 6, 8],
        header=None,
        names=["Chromosome", "Feature", "Start", "End", "Score", "Strand", "Attribute"],
        dtype=dtypes,
        chunksize=int(1e5),
        skiprows=skiprows if skiprows is not None else False,
        nrows=nrows,
    )

    dfs = []
    for df in df_iter:
        if sum(df.Score == ".") == len(df):
            cols_to_concat = ["Chromosome", "Start", "End", "Strand", "Feature"]
        else:
            cols_to_concat = ["Chromosome", "Start", "End", "Strand", "Feature", "Score"]

        extract = _fetch_gene_transcript_exon_id(df.Attribute)
        extract.columns = pd.Index(["gene_id", "group_by", "exon_number", "exon_id"])

        extract.exon_number = extract.exon_number.astype(float)

        extract = extract.set_index(df.index)
        _df = pd.concat([df[cols_to_concat], extract], axis=1, sort=False)

        dfs.append(_df)

    df = pd.concat(dfs, sort=False)

    df.loc[:, "Start"] = df.Start - 1

    return ensure_pyranges(df)


def to_rows_gff3(anno: pd.Series) -> pd.DataFrame:
    """Parse GFF3 attribute column into a dataframe of attribute columns."""
    rowdicts = [to_keys_and_values(line) for line in list(anno)]

    return pd.DataFrame.from_records(rowdicts).set_index(anno.index)


def to_keys_and_values(line: str) -> dict[str, str]:
    """Parse GFF3 attribute column."""
    return dict(it.split("=") for it in line.rstrip("; ").split(";"))


def read_gff3(
    f: str | Path,
    nrows: int | None = None,
    *,
    full: bool = True,
) -> "PyRanges":
    """Read files in the General Feature Format into a PyRanges.

    Parameters
    ----------
    f : str
        Path to GFF file.

    full : bool, default True
        Whether to read and interpret the annotation column.

    nrows : int, default None
        Number of rows to read. Default None, i.e. all.

    Returns
    -------
    PyRanges

    Notes
    -----
    The gff3 format encodes both Start and End as 1-based included.
    PyRanges (and also the DF returned by this function, if as_df=True), instead
    encodes intervals as 0-based, Start included and End excluded.

    See Also
    --------
    pyranges.read_gtf : read files in the Gene Transfer Format

    """
    path = Path(f)
    _skiprows = find_first_data_line_index(path)

    if not full:
        return read_gtf_restricted(path, _skiprows, nrows=nrows)

    dtypes: Mapping = {"Chromosome": "category", "Feature": "category", "Strand": "category"}

    names = ["Chromosome", "Source", "Feature", "Start", "End", "Score", "Strand", "Frame", "Attribute"]

    df_iter = pd.read_csv(
        path,
        comment="#",
        sep="\t",
        header=None,
        names=names,
        dtype=dtypes,
        chunksize=int(1e5),
        skiprows=_skiprows,
        nrows=nrows,
    )

    dfs = []
    for df in df_iter:
        extra = to_rows_gff3(df.Attribute.astype(str))
        _df = df.drop("Attribute", axis=1)
        extra = extra.set_index(_df.index)
        ndf = pd.concat([_df, extra], axis=1, sort=False)
        dfs.append(ndf)

    df = pd.concat(dfs, sort=False)

    df.loc[:, "Start"] = df.Start - 1

    return ensure_pyranges(df)


def read_bigwig(f: str | Path) -> "PyRanges":
    """Read bigwig files into a PyRanges.

    Parameters
    ----------
    f : str
        Path to bw file.

    Returns
    -------
    PyRanges

    Note
    ----
    This function requires the library pyBigWig, it can be installed with pip install pyBigWig

    Examples
    --------
    >>> import pyranges as pr
    >>> path = pr.example_data.files["bigwig.bw"]
    >>> pr.read_bigwig(path)
      index  |      Chromosome    Start      End      Value
      int64  |          object    int64    int64    float64
    -------  ---  ------------  -------  -------  ---------
          0  |               1        0        1        0.1
          1  |               1        1        2        0.2
          2  |               1        2        3        0.3
          3  |               1      100      150        1.4
          4  |               1      150      151        1.5
          5  |              10      200      300        2
    PyRanges with 6 rows, 4 columns, and 1 index columns.
    Contains 2 chromosomes.


    """
    try:
        import pyBigWig  # type: ignore[import]
    except ModuleNotFoundError:
        LOGGER.exception(
            "pyBigWig must be installed to read bigwigs. Use `pip install pyBigWig` to install it.",
        )
        sys.exit(1)

    path = Path(f)
    bw = pyBigWig.open(str(path))

    size = int(1e5)
    chromosomes = bw.chroms()

    dfs = {}

    for chromosome in natsorted(chromosomes):
        outstarts = []
        outends = []
        outvalues = []

        length = chromosomes[chromosome]

        starts = list(range(0, length, size))
        ends = list(range(size, length + size, size))
        ends[-1] = length
        for start, end in zip(starts, ends, strict=True):
            intervals = bw.intervals(chromosome, start, end)
            if intervals is not None:
                for s, e, v in intervals:
                    outstarts.append(s)
                    outends.append(e)
                    outvalues.append(v)

        outstarts = pd.Series(outstarts)
        outends = pd.Series(outends)
        outvalues = pd.Series(outvalues)
        dfs[chromosome] = pd.DataFrame(
            {
                "Chromosome": chromosome,
                "Start": outstarts,
                "End": outends,
                "Value": outvalues,
            },
        )

    return ensure_pyranges(pd.concat(dfs).reset_index(drop=True))
